#pragma warn-

#include <Mega32.h> // Standard defines
#include <delay.h>
#include <stdlib.h>
#include <stdio.h> 

#define XSIZE  319
// make height smaller to fit in text at bottom
#define YSIZE  190

#define XMAXSIZE 2552
#define YMAXSIZE 1592

#define begin {
#define end }

// define button functions
#define MOVE_RIGHT 4
#define MOVE_DOWN  8
#define MOVE_UP    16
#define MOVE_LEFT  32
#define CENTER     64
#define ZOOM       128

// include all the lcd code
unsigned char lcd_buffer[40]; // Buffer for strings
#include "lcd.h"
#include "dflash.h"

#define t0 16

unsigned char time0;

//RXC ISR variables
unsigned char r_busy; //recieve ISR is running
unsigned char r_index; //current string index
unsigned char r_string_size = 0;
unsigned char r_buffer[120]; //input string
unsigned char r_ready; //flag for receive done
unsigned char r_char; //current character
unsigned char sec_index = 0;

//USART parsed buffer variables
unsigned char time[8];   
//unsigned char timeStamp[8];
unsigned char date[6];
unsigned char latit[8];
unsigned char longit[9];
unsigned char last_char;
unsigned char hour, minute, sec, month, yr, latitDirec, longDirec; 
unsigned int day;  

char zoomLevel;
// 0 : Maximum in
// 3 : Full Map

//miniman at location of user
unsigned char manOnOff = 0;
flash char miniMan[7][5] = {{0,1,1,1,0},
			    {0,1,1,1,0},
			    {0,0,1,0,0},
			    {1,1,1,1,1},
			    {0,0,1,0,0},
			    {0,1,0,1,0},
			    {1,0,0,0,1}};
int i,j, flag, page;
char keep_it_rolling = 0;
// page lookup table.  page number of the start of every 100th x1 series
int pageTable[] = {0, 15, 33, 63, 82, 111, 141, 174, 208, 235, 267, 300, 327, 366, 389, 410, 428, 442, 456, 468, 475}; 

int x1,y1,x2,y2;

// ---SPI Pinout
// PB5 (MOSI)
// PB7 (SCK)
// PB4 (CS)
// PB6 (MISO)
   
//latitude/longitude variables
int latit_degree, long_degree, long_min, latit_min;
float latit_sec, long_sec;
//comma count, used in reading nmea
unsigned char comma;


//screen edges (in map pixels)
int xMin, xMax, yMin, yMax;
//current positions (in map pixels)
int xCurrent;
int yCurrent;
//where the window is centered (map pixels)
int xCenter;
int yCenter;
//miniman locations (window pixels)
int xPic;
int yPic;

//button pushing code
char push_flag = 0;
char message_start = 0;
char button_held = 0;
char maybe_held = 0;
//test to see if /any/ button is pushed
//PORTD.0 and 1 are used for UART.
#define is_a_button_pushed() ~(PIND | 0b00000011) 

void process_buttons();
void decodeNMEA(); // Parses the information in the USART buffer
void initialize(); // all init code
void drawMap(); // draws the map
void drawMan(char action); // draws THE MAN.  pass in: on, off, or toggle
void computePosition(); // compute pixels based on lat and long

//**********************************************************
//timer 0 compare ISR
interrupt [TIM0_COMP] void timer0_compare(void) {
  //Decrements the time if it is not already zero
  //Runs every 16 ms
  if (time0>0) --time0;
  if (!(time0%10)) {
       if(!is_a_button_pushed()) {
       		//no button is pushed, so clear the flags
       		push_flag = 0;
       		button_held = 0;
       }
       else {
       		if(!button_held && push_flag) {
	 			//it was down before, so process the push
    	        process_buttons();
		        button_held = 1;
        	}
	        push_flag = 1;
       }
  }
}

//**********************************************************
//UART character-ready ISR
interrupt[USART_RXC] void uart_rec(void)
{
  r_char=UDR; //get a char
  UDR=r_char; //then print it  
  if(r_char == '$') message_start = 1;
  if(message_start) {
	//build the input string
  	if (r_char != '\r') r_buffer[r_index++]=r_char;      
	else {
      putchar('\n'); //use putchar to avoid overwrite  
	  r_buffer[r_index]=0x00; //zero terminate
      r_busy=0; //and clean up
      r_ready=1; //signal cmd processor
	  UCSRB.7=0; //stop rec ISR
	  r_string_size = r_index;
      sec_index = 0; // reset index 
      message_start = 0;
    }
  }

}


void DrawIntro(void) {
	unsigned char i,j;
	WriteCommand(CursorRight); // Cursor direction right
}


void main(void) {
	initialize();
	computePosition();
	xCenter = xCurrent;
	yCenter = yCurrent;
	drawMap();
	drawMan(1); //turn the man on

   while(1){
   		if(time0 == 0){   		
			drawMan(2);
    		if(sec_index < r_string_size) { 
      	 		decodeNMEA();
	   		}
   		time0 = t0;
   		}
   } 

}

void initialize() {
	//--Initialization for USART
	// UBRRH/L = (16Mhz/(16*Baud)) - 1   
	// UBRRH = 0x03; //baud rate of 1200
	// UBRRL = 0x40;
	UBRRL = 207; // Baud rate of 4800,
	// UBRRL = 103; // Baud rate of 9600
   
	UCSRB = 0b10011000; //hex 0x98

	//Buttons
	DDRD |= 0b00000000; //input
   
	//set up timer 0
	//to be used to check USART Buffer
	TIMSK = 2; //enable timer 0
	OCR0 = 250; //compare match value
	TCCR0=0b00001101; //prescaler set to 1024
	time0 = t0;
   
	SPCR  = 0b00000000;

	latit_min = 26;
	latit_sec = 0.711;
	long_min = 28;
	long_sec = 0.939;
	
	zoomLevel = 1;
   
#asm
	sei
#endasm
   
	InitLCD(); // Initialize the LCD  
	DFlashInit();
}

void drawMap() {   
    // are either of the vertices in the box?
    unsigned char inBox;
    //bit 0 = X1
    //bit 1 = Y1
    //bit 2 = X2
    //bit 3 = Y2
    int dx;
    int dy;

    ClearLCD();
	//turn the man off while we clear the screen
    manOnOff = 0;
    

	// calculate window positions based on where the center of the screen is
    xMin = (xCenter - (unsigned int) (((unsigned int) (1 << zoomLevel) * XSIZE)>> 1));
    yMin = (yCenter - (unsigned int) (((unsigned int) (1 << zoomLevel) * YSIZE)>> 1));

   
	// if either maxes or mins are off the map, we must 
	// recenter the viewing window
	if((int)(xMin) < 0) {
   		xMin = 0;
   		xCenter = ((unsigned int) (1 << zoomLevel) * XSIZE)>> 1;
	}
	if((int)(yMin) < 0) {
		yMin = 0;
		yCenter = ((unsigned int) (1 << zoomLevel) * YSIZE)>> 1;
	}
   
   xMax = xMin + (unsigned int) (1 << zoomLevel) * XSIZE;
   yMax = yMin + (unsigned int) (1 << zoomLevel) * YSIZE;

   if(xMax > XMAXSIZE) {
   	xMin = xMin - (xMax - XMAXSIZE);
   	xMax = XMAXSIZE;
   }
   if(yMax > YMAXSIZE) {
   	yMin = yMin - (yMax - YMAXSIZE);
   	yMax = YMAXSIZE;
   }

   page = pageTable[xMin / 100];
   
   i = 0;
   page = 0;
   keep_it_rolling = 1;

   while(keep_it_rolling) {
		for(i=0; i<33; ++i) {
    		x1 = (unsigned int)ReadFlashDF(page, (i<<3));
    		x1 |= ((unsigned int)(ReadFlashDF(page, (i<<3)+1))) << 8;
    		y1 = (unsigned int)ReadFlashDF(page, (i<<3)+2);
    		y1 |= ((unsigned int)(ReadFlashDF(page, (i<<3)+3))) << 8;
    		x2 = (unsigned int)ReadFlashDF(page, (i<<3)+4);
    		x2 |= ((unsigned int)(ReadFlashDF(page, (i<<3)+5))) << 8;
    		y2 = (unsigned int)ReadFlashDF(page, (i<<3)+6);
    		y2 |= ((unsigned int)(ReadFlashDF(page, (i<<3)+7))) << 8;
     		
            inBox = 0;

			if(x1 > xMax || !(x1|y1)) {
   				keep_it_rolling = 0;
   				break;
   			}
   			else {
   				if(x1 >= xMin) inBox = 0b1;
				if(y1 >= yMin && y1 <= yMax) inBox |= 0b10;
				//coord 2
				if(x2 >= xMin && x2 <= xMax) inBox |= 0b100;
				if(y2 >= yMin && y2 <= yMax) inBox |= 0b1000;

				if(((inBox & 0b1) && (inBox & 0b10)) || ((inBox & 0b100) && (inBox & 0b1000))) {
				
	                //adjust whichever coordinate is not inside the box to fit in the window
    	            //X1 is out, Y1, X2, Y2 are in
        	        if(!(inBox & 0b1)) {
	                    //if both coords of the vertex are out of bounds, this is a much harder problem
    	                if(!inBox & 0b10) break;
        	            dx = x2 - x1;
	                    dy = y2 - y1;
    	                y1 = y2 + dy / (dx / (xMin-x1));
        	            x1 = xMin;
	                }
    	            //Y1 is out, X1, X2, Y2 are in
        	        else if(!(inBox & 0b10)) {
            	        //if both coords of the vertex are out of bounds, this is a much harder problem
	                    if(!inBox & 0b1) break;
	                    if(y1 < yMin) {
	    	                dx = x2 - x1;
    	    	            dy = y2 - y1;
	    	                x1 = x2 - dx / (dy / (yMin - y1));
    	    	            y1 = yMin;
    					}
    					else {
		                    dx = x2 - x1;
    		                dy = y2 - y1;
        		            x1 = (x1 + dx / (dy / (y1 - yMax)));
            		        y1 = yMax;
    					}
        	        }
	                //X2 is out, X1, Y1, Y2 are in
    	            else if(!(inBox & 0b100)) {
        	            //if both coords of the vertex are out of bounds, this is a much harder problem
            	        if(!inBox & 0b1000) break;
        	            dx = x2 - x1;
	                    dy = y2 - y1;
    	                y2 = (y2 - dy / (dx / (x2 - xMax)));
        	            x2 = xMax;
	                }
    	            //Y2 is out, X1, Y1, X2 are in
        	        else if(!(inBox & 0b1000)) {
	                    //if both coords of the vertex are out of bounds, this is a much harder problem
    	                if(!inBox & 0b100) break;
    	                if(y2 > yMax) {
		                    dx = x2 - x1;
    		                dy = y2 - y1;
        		            x2 = (x2 - dx / (dy / (y2 - yMax)));
            		        y2 = yMax;
         				}
         				else {
	    	                dx = x2 - x1;
    	    	            dy = y2 - y1;
	    	                x2 = x1 + dx / (dy / (yMin - y2));
    	    	            y2 = yMin;
         				}
	                }
					x1 = (x1-xMin) >> zoomLevel;
   					x2 = (x2-xMin) >> zoomLevel;
	   				y1 = (y1-yMin) >> zoomLevel;
   					y2 = (y2-yMin) >> zoomLevel;
   					lcd_line(x1,y1,x2,y2,1); 
				}
			}
		}
   page++;
   }

}

// draw man
void drawMan(char action) {
	if((action == 0 && manOnOff) || (action == 1 && !manOnOff) || (action == 2)) {
		if(yCurrent < yMax && yCurrent > yMin && xCurrent < xMax && xCurrent > xMin) {
			xPic = (xCurrent-xMin) >> zoomLevel;
			yPic = (yCurrent-yMin) >> zoomLevel;
			for(i=0; i<7; ++i) {
				for(j=0; j<5; ++j) {
					if(miniMan[i][j]) lcd_pt(xPic - 2 + j, yPic - 3 + i, 2);
				}
			}
		}
		manOnOff = ~manOnOff;
	}
}


void decodeNMEA() {

  int i;
  unsigned char current_time = 0;
  unsigned char current_lat = 0;
  unsigned char current_long = 0;
  unsigned char current_date = 0;
  unsigned char comma = 0;   

  //sample buffer info
  //$GPRMC,175402,A,4226.4300,N,07629.1980,W,005.0,315.0,250404,000.0,W*79
  //$GPRMC,161955,A,0000.0350,N,00000.0350,W,005.0,315.0,250404,000.0,W*7C
  //comma 1      2 3         4 5          6 7     8     9      10    11  
  sec_index = 0;
  comma=0;

  last_char = r_buffer[sec_index++];
 
  while(comma < 1) { //increment the index until the latitude
    last_char = r_buffer[sec_index++];
    if(last_char == ',') comma++; 
  }   
  last_char = r_buffer[sec_index++];
  while(last_char != ',') { //get the time
    time[current_time] = last_char;
    current_time++;
    if(current_time == 2 || current_time==5) time[current_time++] = ':';
    last_char = r_buffer[sec_index++];
  } 
  flag =1;    
  comma++;   // comma = 2

  while(comma < 3) { //increment the index until the latitude
    last_char = r_buffer[sec_index++];
    if(last_char == ',') comma++; 
  }


  //last_char = r_buffer[sec_index++];
  last_char = r_buffer[sec_index++];
     
  while(last_char != ',') { 
    latit[current_lat] = last_char;

    current_lat++;
    last_char = r_buffer[sec_index++];
  }                                   
     
  latit_degree = (latit[0]-'0')*10 + (latit[1]-'0');
  latit_min = (latit[2]-'0')*10 + (latit[3]-'0');
  latit_sec = (latit[5]-'0')*.1 + (latit[6]-'0')*.01 + (latit[7]-'0')*.001;// + (latit[8]-'0')*.0001);
      
                      
  comma++; //comma before latitude direction 
  latitDirec = r_buffer[sec_index++]; 

  while(comma < 5) { //increment the index until the latitude
    last_char = r_buffer[sec_index++];
    if(last_char == ',') comma++; 
  }   

  last_char = r_buffer[sec_index++];
  while(last_char != ',') {
    longit[current_long] = last_char;
    current_long++;
    last_char = r_buffer[sec_index++];
  }

  long_degree = (longit[0]-'0')*100 + (longit[1]-'0')*10 + (longit[2]-'0');
  long_min = (longit[3]-'0')*10 + (longit[4]-'0');
  long_sec = (longit[6]-'0')*0.1 + (longit[7]-'0')*0.01 + (longit[8]-'0')*.001;// + (longit[9]-'0')*.0001;
     
     
  comma++; //comma before longitude direction
  longDirec = r_buffer[sec_index++];
  if(r_buffer[sec_index++] == ',') comma++; //comma after   
  //comma = 7
  
  while(comma < 9) { //increment the index until the date
    last_char = r_buffer[sec_index++];
    if(last_char == ',') comma++; 
  }   
  
  last_char = r_buffer[sec_index++];
  while(last_char != ',') {
    date[current_date] = last_char;
    current_date++;
    last_char = r_buffer[sec_index++];
  }
  comma++;                
  
  // day = (date[0] - '0')*10 + (date[1] - '0');
  // month = (date[2] - '0')*10 + (date[3] - '0');
  // yr = (date[4] - '0')*10 + (date[5] - '0');
  
  MoveCursor(960);
  // print out the time, but not the seconds
  WriteCommand(MemoryWrite);
  for(i=0; i<5; i++) WriteData(time[i]);
  // print out the latitude
  WriteData(' ');
  for(i=0; i<9; i++) WriteData(latit[i]);
  WriteData(latitDirec);
  WriteData(' ');
  // print out the longitude
  for(i=0; i<10; i++) WriteData(longit[i]);
  WriteData(longDirec);
  // print out the date
  //  for(i=0; i<6; i++) WriteData(date[i]);
                                            
  WriteData(' ');
  WriteData(date[2]);
  WriteData(date[3]);
  WriteData('/');
  WriteData(date[0]);   
  WriteData(date[1]);
  WriteData('/');
  WriteData(date[4]);
  WriteData(date[5]);
  
  computePosition();

  r_index = 0;
  r_busy=0; //and clean up
  r_ready=1; //signal cmd processor
  r_string_size = 0;        
  UCSRB.7=1; //start rec ISR
  
}

void computePosition() {
	drawMan(0);
	xCurrent = (int) (-1048.3 * (long_min + long_sec) + 30963);
 	yCurrent = (int) (-1464.0 * (latit_min + latit_sec) + 40244);
}

void process_buttons() {
 	switch(~(PIND | 0b11)) {
 		case MOVE_RIGHT:
			xCenter += 200 * (1 << zoomLevel);
			break;
 		case MOVE_LEFT:
 			xCenter -= 200 * (1 << zoomLevel);
			break;
 		case MOVE_UP:
 			yCenter = yCenter - 100 * (1 << zoomLevel);
			break;
 		case MOVE_DOWN:
 			yCenter = yCenter + 100 * (1 << zoomLevel);
			break;
		case ZOOM:
			if(++zoomLevel == 4) zoomLevel = 0;
			//center on zoom as well
		case CENTER:
			xCenter = xCurrent;
			yCenter = yCurrent;
			break;
	}
	drawMap();
}
